//** Modified-Light version of Triangle for TP**// 




#ifndef   	TRIANGLE_H_
# define   	TRIANGLE_H_

#include <V_3D.h>
#include <Segment.h>

#include <vector>
using namespace std;


/// Manipulation class for a Triangle
/** Calculation for embedeed Triangle. */


class Triangle
{
public:

  //*********************************************//
  // CONSTRUCTORS
  //*********************************************//
  
  /// Classical constructor
  Triangle();
  /// Direct constructor
  Triangle(const V_3D& x0,const V_3D& x1,const V_3D& x2);
  /// Copy Constructor
  Triangle(const Triangle& t);
  /// destructor
  ~Triangle();

  //*********************************************//
  // Manipulation
  //*********************************************//
  
  /// set every vertices to 0
  int destroy();

  /// return the unitary normal vector
  V_3D normal() const;

  /// return the area of the triangle
  double area() const;
  
  /// get segment of an edge
  /** k_edge has to be between 0 and 2 */
  Segment get_segment(int k_edge) const;


  //*********************************************//
  // Informations
  //*********************************************//

  /// get barycentric coordinates for vertex in this plane
  /** return 0 and (alpha,beta,gamma) if x is in the plane, 
      and return -1 if x is not in the plane */
  int barycentric_coordinates(const V_3D& x,double *alpha,double *beta,double *gamma) const;

  /// check if a vertex (in plane) is inside the triangle by calculating its barycentric coordinates
  int is_vertex_inside(const V_3D _x) const;


  /// get distance of a point to a triangl
  /** 2 possibilities: 
             1. the closest point is the projection of the point onto the triangle
             2. the closest point is on the boundary lines 

      type=0 if the closest point is in the triangle
      type=1 if the closest point is on the edge
      type=2 if the closest point is on a vertice
  */
  V_3D closest_point(const V_3D& x,int *type) const;

  /// Intersection with the plane of equation <n,x-x0>=0
  /** The plane is defined by its normal n, and a point x0, so its equation is <n,x-x0>=0
      4 possibilities: _ no intersections (type 0) => return void vector
		       _ plane intersect on two distinct points (type 1) => return vector of size 2
		       _ plane intersect only on a vertex (type 2) => return vector of size 1
                       _ plane intersect along an edge of the triangle (type 3 > return the two extreme vertices) => return a vector of size 2

   */
  std::vector <V_3D> plane_intersection(const V_3D& n,const V_3D& x0,int *type) const;
  ///  Intersection with the plane of equation <n,x-x0>=0
  /** Same as plane_intersection function, but with the detailed type intersection for each edge */
  std::vector <V_3D> plane_intersection(const V_3D& n,const V_3D& x0,int *type,int *type_edge_0,int *type_edge_1,int *type_edge_2) const;

  //*********************************************//
  // Operators
  //*********************************************//

  /// direct access operator
  V_3D  operator()(int index) const;
  /// direct access operator
  V_3D& operator()(int index);
  /// direct access operator
  V_3D  operator[](int index) const;
  /// direct access operator
  V_3D  operator[](int index);

  /// copy Triangle
  Triangle& operator=(const Triangle& t);


private:
  
  V_3D x[3];


};

#endif 	    /* !TRIANGLE_H_ */
