//** Modified-Light version of V_4D for TP**// 


#ifndef   	V_4D_H_
# define   	V_4D_H_

#include <stdlib.h>
#include <iostream>
#include <cmath>
#include <string>
#include <errno.h>
#include <stdio.h>

/// The V_4D class, container for vector
/** Container Class for 4D vector **/

class V_4D
{
 public:
  


  //************************************************//
  //************************************************//
  //CONSTRUCTORS
  //************************************************//
  //************************************************//

  /// Classical constructor
  V_4D();
  /// Copy constructor
  V_4D(const V_4D &V_4D);
  /// Direct init
  V_4D(double x,double y,double z,double w);
  /// destructor
  ~V_4D();

  //************************************************//
  //************************************************//
  //ASSIGN
  //************************************************//
  //************************************************//

  /// set all coordinates to 0
  int set_zero();
  /// set [0,0,0,1]
  int set_3d_zero();

  //************************************************//
  //************************************************//
  //INFORMATIONS
  //************************************************//
  //************************************************//

  /// test equality with error epsilon
  int equal(const V_4D& v,double epsilon) const;
  /// test equality (epsilon=10^-5)
  int equal(const V_4D& v) const;


  //************************************************//
  //************************************************//
  //OPERATORS
  //************************************************//
  //************************************************//

  /// get value
  double& operator[](int index);
  /// get value
  double operator[](int index) const;
  /// get value
  double& operator()(int index);
  /// get value
  double operator()(int index) const;


  /// assign
  V_4D& operator=(const V_4D&);
  /// add vector
  V_4D& operator+=(const V_4D&);
  /// divide by scalar
  V_4D& operator/=(const double&);

  /// add vectors
  friend V_4D operator+(const V_4D& v1,const V_4D& v2);
  /// substract vector
  friend V_4D operator-(const V_4D& v1,const V_4D& v2);
  /// multiply by scalar
  friend V_4D operator*(const V_4D& v1,const double& alpha);
  /// multiply by scalar
  friend V_4D operator*(const double& alpha,const V_4D& v1);
  /// divide by scalar
  friend V_4D operator/(const V_4D& v1,const double& alpha);

  /// test equality
  friend bool operator==(const V_4D& v1,const V_4D& v2);
  /// test inequality
  friend bool operator!=(const V_4D& v1,const V_4D& v2);

  /// cout operator
  friend std::ostream& operator << (std::ostream& flux,const V_4D& _v);



 private:

  double V[4];

};

#endif  /* !V_4D_H_h */
