//** Modified-Light version of Connectivity for TP**// 






#include <Connectivity.h>


Connectivity::Connectivity()
{
}

Connectivity::Connectivity(const Connectivity& c)
{
  index_polygon  = c.index_polygon;
  size_polygon   = c.size_polygon;
  polygon_access = c.polygon_access;
  max_index      = c.max_index;
  one_ring       = c.one_ring;
  polygon_ring   = c.polygon_ring;
}

Connectivity::~Connectivity(){destroy();}

int Connectivity::destroy()
{
  index_polygon  . resize(0);
  size_polygon   . resize(0);
  polygon_access . resize(0);
  one_ring       . resize(0);
  polygon_ring   . resize(0);
  max_index=-1;

  return 0;
}

int Connectivity::get_index(int k_index) const
{
  if(k_index<0 || k_index>=size_index())
    {printf("Error in get_index(%d) with size %d in Connectivity\n",k_index,size_index());exit(-1);}
  return index_polygon[k_index];
}
int& Connectivity::get_index(int k_index)
{
  if(k_index<0 || k_index>=size_index())
    {printf("Error in get_index(%d) with size %d in Connectivity\n",k_index,size_index());exit(-1);}
  return index_polygon[k_index];
}
int Connectivity::get_index(int k_polygon,int k_index) const
{
  if(k_polygon<0 || k_polygon>=polygon_number())
    {printf("Error in get_index, k_polygon(%d) is not correct with size %d in Connectivity\n",k_polygon,polygon_number());exit(-1);}
  if(k_index<0 || k_index>=polygon_size(k_polygon))
    {printf("Error in get_index, k_index(%d) is not correct for polygon %d in Connectivity. This polygon has size=%d\n",k_index,k_polygon,polygon_size(k_polygon));exit(-1);}
  return index_polygon[polygon_access[k_polygon]+k_index];
}
int& Connectivity::get_index(int k_polygon,int k_index)
{
  if(k_polygon<0 || k_polygon>=polygon_number())
    {printf("Error in get_index, k_polygon(%d) is not correct with size %d in Connectivity\n",k_polygon,polygon_number());exit(-1);}
  if(k_index<0 || k_index>=polygon_size(k_polygon))
    {printf("Error in get_index, k_index(%d) is not correct for polygon %d in Connectivity. This polygon has size=%d\n",k_index,k_polygon,polygon_size(k_polygon));exit(-1);}
  return index_polygon[polygon_access[k_polygon]+k_index];
}
std::vector <int> Connectivity::get_index_of_polygon(int k_polygon) const
{
  if(k_polygon<0 || k_polygon>=polygon_number())
    {printf("Error in get_index, k_polygon(%d) is not correct with size %d in Connectivity\n",k_polygon,polygon_number());exit(-1);}
  std::vector <int> ind;
  int k_index=0,N_index=polygon_size(k_polygon);
  for(k_index=0;k_index<N_index;k_index++)
    ind.push_back(get_index(k_polygon,k_index));
  return ind;
}
const int* Connectivity::get_index() const {return &index_polygon[0];}






int Connectivity::size_index() const {return index_polygon.size();}
int Connectivity::polygon_number() const {return size_polygon.size();}
int Connectivity::polygon_size(int k_polygon) const
{
  if(k_polygon<0 || k_polygon>=int(size_polygon.size()))
    {printf("Error in polygon_size in Connectivity, k_polygon[%d] with size[%d]\n",k_polygon,size_polygon.size());exit(-1);}
  return size_polygon[k_polygon];
}




int Connectivity::get_index_triangulated(int k_index) const
{
  if(k_index<0 || k_index>=int(triangulated_index.size()))
    {printf("Error in get_index_triangulated in Connectivity, k_index[%d] with size[%d]\n",k_index,triangulated_index.size());exit(-1);}
  return triangulated_index[k_index];
}
int& Connectivity::get_index_triangulated(int k_index)
{
  if(k_index<0 || k_index>=int(triangulated_index.size()))
    {printf("Error in get_index_triangulated in Connectivity, k_index[%d] with size[%d]\n",k_index,triangulated_index.size());exit(-1);}
  return triangulated_index[k_index];
}
int Connectivity::get_index_triangulated(int k_triangle,int k_index) const
{
  if(k_triangle<0 || 3*k_triangle+2>=int(triangulated_index.size()))
    {printf("Error in get_index_triangulated in Connectivity, k_triangle[%d] with size[%d]\n",k_triangle,k_index);exit(-1);}
  if(k_index<0 || k_index>=3)
    {printf("Error in get_index_triangulated in Connectivity, k_index[%d] should be between 0 and 2\n",k_index);exit(-1);}
  return triangulated_index[3*k_triangle+k_index];
}
int& Connectivity::get_index_triangulated(int k_triangle,int k_index)
{
 if(k_triangle<0 || 3*k_triangle+2>=int(triangulated_index.size()))
    {printf("Error in get_index_triangulated in Connectivity, k_triangle[%d] with size[%d]\n",k_triangle,k_index);exit(-1);}
  if(k_index<0 || k_index>=3)
    {printf("Error in get_index_triangulated in Connectivity, k_index[%d] should be between 0 and 2\n",k_index);exit(-1);}
  return triangulated_index[3*k_triangle+k_index];
}
const int* Connectivity::get_index_triangulated() const
{return &triangulated_index[0];}
int Connectivity::get_triangle_number() const
{
  if(triangulated_index.size()%3!=0)
    {printf("Error size of triangulated_index [%d] is not multiple of 3\n",triangulated_index.size());exit(-1);}
  return triangulated_index.size()/3;
}




int Connectivity::build_size_and_polygon_access()
{
  int k_index=0,N_index=size_index();
  int current_size=0,k_polygon=0;

  size_polygon.resize(0);
  polygon_access.resize(0);

  //the first polygon start at index 0
  polygon_access.push_back(0);
  for(k_index=0;k_index<N_index;k_index++)
    {
      if(index_polygon[k_index]==-1)
	{
	  if(current_size<2)
	    {printf("Error in build_size_and_polygon_access in Connectivity, polygon_size[%d] is less than 2 at index[%d]\n",current_size,k_index);exit(-1);}
	  size_polygon.push_back(current_size);
	  current_size=0;
	  k_polygon++;
	}
      else
	current_size++;
      if(k_index>0 && index_polygon[k_index-1]==-1)
	polygon_access.push_back(k_index);

    }
  if(index_polygon[N_index-1]!=-1)
    {printf("Error index_polygon of size(%d) does not end with -1 [%d]\n",index_polygon.size(),index_polygon[N_index-1]);exit(-1);}
  return 0;
}

int Connectivity::build_triangulated_index()
{
  triangulated_index.resize(0);
  int k_index=0,N_index;
  int k_polygon=0,N_polygon=polygon_number();

  int index0,index1,index2;
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    {
      N_index=polygon_size(k_polygon);
      index0=get_index(k_polygon,0);
      index1=get_index(k_polygon,1);
      for(k_index=2;k_index<N_index;k_index++)
	{
	  index2=get_index(k_polygon,k_index);
	  
	  //add the current triangle
	  triangulated_index.push_back(index0);
	  triangulated_index.push_back(index1);
	  triangulated_index.push_back(index2);

	  //reorganize the current triangle
	  //do triangle strip
	  index1=index2;
	}
      
    }
  return 0;
}

int Connectivity::find_max_index()
{
  max_index=-2;
  int k_index=0,N_index=size_index();
  for(k_index=0;k_index<N_index;k_index++)
    if(max_index<get_index(k_index))
      max_index=get_index(k_index);
  return 0;
}

int Connectivity::build_all_index()
{
  int ok=0;
  ok += build_size_and_polygon_access();
  ok += build_triangulated_index();
  ok += find_max_index();
  ok += build_one_ring();
  return ok;
}


int Connectivity::add_polygon(int index_0,int index_1,int index_2)
{
  std::vector <int> polygon;
  polygon.push_back(index_0);
  polygon.push_back(index_1);
  polygon.push_back(index_2);
  return add_polygon(polygon);
}
int Connectivity::add_polygon(int index_0,int index_1,int index_2,int index_3)
{
  std::vector <int> polygon;
  polygon.push_back(index_0);
  polygon.push_back(index_1);
  polygon.push_back(index_2);
  polygon.push_back(index_3);
  return add_polygon(polygon);
}
int Connectivity::add_polygon(const int_vector& polygon)
{return add_polygon(polygon.to_vector());}
int Connectivity::add_polygon(const std::vector <int>& polygon)
{
  int current_size=polygon.size();
  if(current_size<3)
    {printf("Error polygon size [%d] in add_polygon<3\n",current_size);exit(-1);}


  //test if the given polygon is correct
  int N_poly=polygon.size();
  std::vector <int> sorted_poly=polygon;
  std::sort(sorted_poly.begin(),sorted_poly.end());
  for(int k=0;k<N_poly-1;k++)
    if(sorted_poly[k]==sorted_poly[k+1])
      {printf("Error polygon given in add_polygon in Connectivity is not correct, two index are the same=degenerated polygon [%d]\n",sorted_poly[k]);}




  int k=0;
  for(k=0;k<current_size;k++)
    index_polygon.push_back(polygon[k]);

  //add the final -1
  index_polygon.push_back(-1);
  
  return 0;
}

int Connectivity::add_index(int index)
{
  index_polygon.push_back(index);
  return 0;
}

int Connectivity::set_polygon(int k_polygon,const std::vector <int>& polygon)
{

  if(k_polygon<0 || k_polygon>=int(polygon_access.size()))
    {printf("Error k_polygon[%d] is not correct in set_polygon in Connectivity (number of polygon: %d)\n",k_polygon,polygon_access.size());exit(-1);}


  std::vector <int> index_polygon_2;
  int k_access0=0;k_access0=polygon_access[k_polygon];
  int k_access1=-1;if(k_polygon<int(polygon_access.size())-1){k_access1=polygon_access[k_polygon+1];}
  int k_index=0;

  //copy of the first part of the polygon
  for(k_index=0;k_index<k_access0;k_index++)
    index_polygon_2.push_back(index_polygon[k_index]);

  //then replace the polygon
  int N_polygon=0;N_polygon=polygon.size();
  if(N_polygon<3){printf("Error polygon size [%d] is <3 in set_polygon(%d) in Connectivity\n",N_polygon,k_polygon);exit(-1);}
  for(k_index=0;k_index<N_polygon;k_index++)
    index_polygon_2.push_back(polygon[k_index]);
  //add the final -1
  index_polygon_2.push_back(-1);

  //finally add the other one
  if(k_access1!=-1)
    {
      for(k_index=k_access1;k_index<int(index_polygon.size());k_index++)
	index_polygon_2.push_back(index_polygon[k_index]);
    }

  //then copy
  index_polygon=index_polygon_2;

  //recalculate everything
  build_all_index();

  return 0;
}


int Connectivity::set_index(int k_polygon,int k_index,int index)
{

  if(k_polygon<0 || k_polygon>=int(polygon_access.size()))
    {printf("Error k_polygon[%d] is not correct in set_index in Connectivity (number of polygon: %d)\n",k_polygon,polygon_access.size());exit(-1);}

  if(k_index<0 || k_index>=polygon_size(k_polygon))
    {printf("Error k_index[%d] is not correct with size[%d] of polygon[%d]\n",k_index,polygon_size(k_polygon),k_polygon);exit(-1);}

  int current_index=polygon_access[k_polygon]+k_index;
  index_polygon[current_index]=index;
  return 0;
}


int Connectivity::delete_polygon(int k_polygon)
{
  if(k_polygon<0 || k_polygon>=int(polygon_access.size()))
    {printf("Error k_polygon[%d] is not correct in set_index in Connectivity (number of polygon: %d)\n",k_polygon,polygon_access.size());exit(-1);}

  int k_access0=polygon_access[k_polygon];
  int k_access1=-1;if(k_polygon<int(polygon_access.size())-1){k_access1=polygon_access[k_polygon+1];}
  int k_index=0;

  //copy until reach the polygon
  std::vector <int> index_polygon_2;
  for(k_index=0;k_index<k_access0;k_index++)
    index_polygon_2.push_back(index_polygon[k_index]);
  
  //copy the last part
  if(k_access1!=-1)
    {
      for(k_index=k_access1;k_index<int(index_polygon.size());k_index++)
	index_polygon_2.push_back(index_polygon[k_index]);
    }
  
  //copy
  index_polygon=index_polygon_2;

  //rebuild_index
  build_all_index();

  return 0;
}

int Connectivity::delete_given_polygon(const std::vector <int> polygon)
{
  int k_poly_to_delete=find_polygon(polygon);
  if(k_poly_to_delete!=-1)
    return delete_polygon(k_poly_to_delete);

  //did not find the polygon
  return -1;
}







int Connectivity::load_off_file(const char* filename)
{std::string file=filename;return load_off_file(file);}
int Connectivity::load_off_file(const std::string& filename)
{
  destroy();

  #define SIZE_BUFFER 2048
  char buffer[SIZE_BUFFER]={'\0'};

  FILE *fid=NULL;
  fid=fopen(filename.data(),"r");
  if(fid==NULL)
    {printf("ERROR loading %s in load_off_file in Connectivity\n",filename.data());exit(-1);}


  //find OFF or NOFF
  while(strcmp(buffer,"OFF")!=0 && strcmp(buffer+1,"OFF")!=0)
    {
      if(fscanf(fid,"%s",buffer)==EOF)
	{printf("ERROR file %s is not off file (in Connectivity)\n",filename.data());exit(-1);}
      if(buffer[0]=='#')//comments
	fgets(buffer,sizeof(buffer),fid);
    }

  //pass comments
  fscanf(fid,"%s",buffer);
  while(buffer[0]=='#')//comments
    {
      fgets(buffer,sizeof(buffer),fid);
      fscanf(fid,"%s",buffer);
    }

  //read Number of polygon
  int N_vertex=0,N_polygon=0;
  N_vertex=atoi(buffer);//N_vertex
  fscanf(fid,"%d",&N_polygon);//N_polygon
  fgets(buffer,sizeof(buffer),fid);//read line

  //read vertices
  int k_vertex=0;
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    {
      fgets(buffer,sizeof(buffer),fid);//read line
      if(buffer[0]=='#')//comments does not count in vertices number
	k_vertex--;
    }

  //read polygon
  int size_poly=0;
  int k_polygon=0;
  int k_current_polygon=0;
  int temp=0;
  std::vector <int> v_poly;
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    {
       if(fscanf(fid,"%d",&size_poly)!=1)
	{
	  fscanf(fid,"%s",buffer);
	  if(buffer[0]!='#')
	    {printf("ERROR loading polygons in load_off_file %s in Connectivity\n",filename.data());exit(-1);}
	  else//read line
	    {
	      fgets(buffer,sizeof(buffer),fid);
	      k_polygon--;
	    }
	}
      else
	{
	  v_poly.resize(0);
	  for(k_current_polygon=0;k_current_polygon<size_poly;k_current_polygon++)
	    {
	      fscanf(fid,"%d",&temp);
	      v_poly.push_back(temp);
	    }
	  add_polygon(v_poly);
	}
    }
    
  if(fclose(fid)!=0)
    {printf("ERROR closing %s in load_off_file in Connectivity\n",filename.data());exit(-1);}


  //build the index
  build_all_index();

  return 0;
}

int Connectivity::write_off_type(const char* filename) const
{std::string file=filename;return write_off_type(file);}

int Connectivity::write_off_type(const std::string& filename) const
{
  FILE *fid=NULL;
  fid=fopen(filename.data(),"a");
  if(fid==NULL)
    {printf("ERROR loading %s in write_off_type in Connectivity\n",filename.data());exit(-1);}

  int k_polygon=0,N_polygon=polygon_number();
  int current_size=0,k_index;
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    {
      current_size=polygon_size(k_polygon);

      //write the size of the polygon
      fprintf(fid,"%d ",current_size);
      //read the polygon
      for(k_index=0;k_index<current_size;k_index++)
	fprintf(fid,"%d ",get_index(k_polygon,k_index));
      fprintf(fid,"\n");
    }

  if(fclose(fid)!=0)
    {printf("ERROR closing %s in write_off_type in Connectivity\n",filename.data());exit(-1);}

  return 0;

}






int Connectivity::operator()(int k_index) const
{
  if(k_index<0 || k_index>=size_index())
    {printf("Error in operator()(%d) with size %d in Connectivity\n",k_index,size_index());exit(-1);}
  return index_polygon[k_index];
}
int Connectivity::operator[](int k_index) const
{
  if(k_index<0 || k_index>=size_index())
    {printf("Error in operator[](%d) with size %d in Connectivity\n",k_index,size_index());exit(-1);}
  return index_polygon[k_index];
}
int& Connectivity::operator()(int k_index)
{
  if(k_index<0 || k_index>=size_index())
    {printf("Error in operator()(%d) with size %d in Connectivity\n",k_index,size_index());exit(-1);}
  return index_polygon[k_index];
}
int& Connectivity::operator[](int k_index)
{
  if(k_index<0 || k_index>=size_index())
    {printf("Error in operator[](%d) with size %d in Connectivity\n",k_index,size_index());exit(-1);}
  return index_polygon[k_index];
}
int Connectivity::operator()(int k_polygon,int k_index) const
{
  if(k_polygon<0 || k_polygon>=polygon_number())
    {printf("Error in operator(), k_polygon(%d) is not correct with size %d in Connectivity\n",k_polygon,polygon_number());exit(-1);}
  if(k_index<0 || k_index>=polygon_size(k_polygon))
    {printf("Error in operator(), k_index(%d) is not correct for polygon %d in Connectivity. This polygon has size=%d\n",k_index,k_polygon,polygon_size(k_polygon));exit(-1);}
    return index_polygon[polygon_access[k_polygon]+k_index];
}
int& Connectivity::operator()(int k_polygon,int k_index)
{
  if(k_polygon<0 || k_polygon>=polygon_number())
    {printf("Error in operator(), k_polygon(%d) is not correct with size %d in Connectivity\n",k_polygon,polygon_number());exit(-1);}
  if(k_index<0 || k_index>=polygon_size(k_polygon))
    {printf("Error in operator(), k_index(%d) is not correct for polygon %d in Connectivity. This polygon has size=%d\n",k_index,k_polygon,polygon_size(k_polygon));exit(-1);}
    return index_polygon[polygon_access[k_polygon]+k_index];
}


int Connectivity::find_polygon(const std::vector <int>& polygon) const
{
  int N_poly=polygon.size();
  
  //test if the given polygon is correct
  std::vector <int> sorted_poly=polygon;
  std::sort(sorted_poly.begin(),sorted_poly.end());
  for(int k=0;k<N_poly-1;k++)
    if(sorted_poly[k]==sorted_poly[k+1])
      {printf("Error polygon given in find_polygon in Connectivity is not correct, two index are the same=degenerated polygon [%d]\n",sorted_poly[k]);exit(-1);}
  
  int N_polygon=polygon_number();
  std::vector <int> compared_polygon;
  int k_polygon=0;
  int count=0,k_poly=0,k_to_compare=0;

  //loop on every polygon of the mesh
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    {
      //get polygon to compare
      compared_polygon=get_index_of_polygon(k_polygon);


      count=0;
      //test if size are equal
      if(N_poly==int(compared_polygon.size()))
	{
	  //test for every index of the given polygon
	  for(k_poly=0;k_poly<N_poly;k_poly++)
	    //test for every index of the current compared polygon
	    for(k_to_compare=0;k_to_compare<int(compared_polygon.size());k_to_compare++)
	      //if one index is the same
	      if(polygon[k_poly]==compared_polygon[k_to_compare])
		count++;
	}
      if(count==N_poly)//find the polygon
	return k_polygon;
    }
  return -1;
}


const Connectivity& Connectivity::operator=(const Connectivity& c)
{
  index_polygon      = c.index_polygon;
  size_polygon       = c.size_polygon;
  polygon_access     = c.polygon_access;
  triangulated_index = c.triangulated_index;
  one_ring           = c.one_ring;
  polygon_ring       = c.polygon_ring;

  return (*this);
}

Connectivity operator+(const Connectivity& c1,const Connectivity& c2)
{
  Connectivity c;
  int N_polygon=0;
  int k_polygon=0;

  N_polygon=c1.polygon_number();
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    c.add_polygon(c1.get_index_of_polygon(k_polygon));

  N_polygon=c2.polygon_number();
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    c.add_polygon(c2.get_index_of_polygon(k_polygon));

  c.build_all_index();

  return c;
}

Connectivity operator+(const Connectivity& _c,const std::vector<int> & v)
{
  Connectivity c;

  int N_polygon=0;
  int k_polygon=0;

  N_polygon=_c.polygon_number();
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    c.add_polygon(_c.get_index_of_polygon(k_polygon));

  c.add_polygon(v);
  c.build_all_index();

  return c;
}


Connectivity operator+(const std::vector<int> & v,const Connectivity& _c)
{
  Connectivity c;
  c = _c+v;
  return c;
}


Connectivity operator-(const Connectivity& c1,const Connectivity& c2)
{
  Connectivity c;
  c=c1;

  int k=0;
  int N_polygon=c2.polygon_number();
  for(k=0;k<N_polygon;k++)
    c.delete_given_polygon(c2.get_index_of_polygon(k));
  
  return c;
}

Connectivity operator-(const Connectivity& c1,const std::vector<int>& c2)
{
  Connectivity c=c1;
  c.delete_given_polygon(c2);
  return c;
}


bool operator==(const Connectivity& c1,const Connectivity& c2)
{

  //must have same dimension first
  if(c1.polygon_number()!=c2.polygon_number())
    return 0;

  int k_polygon=0;
  int N_polygon=c1.polygon_number();
  int k_vertex=0,N_vertex=0;
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    {
      if(c1.polygon_size(k_polygon)!=c2.polygon_size(k_polygon))
	return 0;

      N_vertex=c1.polygon_size(k_polygon);
      for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
	if(c1(k_polygon,k_vertex)!=c2(k_polygon,k_vertex))
	  return 0;
    }
  return 1;
}

bool operator!=(const Connectivity& c1,const Connectivity& c2){return !(c1==c2);}

ostream& operator << (ostream& flux,const Connectivity& c)
{
  int k_polygon=0,N_polygon=c.polygon_number();
  int k_vertex=0,N_vertex=0;
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    {
      N_vertex=c.polygon_size(k_polygon);
      flux<<N_vertex<<" [";
      for(k_vertex=0;k_vertex<N_vertex-1;k_vertex++)
	cout<<c(k_polygon,k_vertex)<<",";
      cout<<c(k_polygon,N_vertex-1)<<"]"<<endl;
    }

  return flux;
}


int Connectivity::vertex_number() const{return max_index+1;}


int Connectivity::build_one_ring()
{

  int k_polygon=0,k_vertex=0,N_vertex=0;
  int k_index=0,k_index_2=0;
  int N_polygon=0;

  if(max_index<=0)
    {printf("Error in build_one_ring, max_index=0\n");exit(-1);}

  one_ring.resize(max_index+1);
  polygon_ring.resize(max_index+1);


  //loop over every vertices
  N_polygon=polygon_number();
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    {
      N_vertex=polygon_size(k_polygon);
      for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
	{
	  k_index = get_index(k_polygon,k_vertex);

	  //add the following vertex
	  k_index_2 = get_index(k_polygon,(k_vertex+1)%N_vertex);
	  one_ring[k_index]   . add_unique(k_index_2);
	  one_ring[k_index_2] . add_unique(k_index);

	  //add the previous vertex	  
	  k_index_2 = get_index(k_polygon,(k_vertex-1)<0?N_vertex-1:k_vertex-1);
	  one_ring[k_index]   . add_unique(k_index_2);
	  one_ring[k_index_2] . add_unique(k_index);

	  // add the polygon_ring
	  polygon_ring[k_index].add_unique(k_polygon);
	}


    }    

  
  return 0;
}

int_vector Connectivity::get_one_ring(int k_polygon,int k_edge) const
{
  int k_index = get_index(k_polygon,k_edge);
  if(k_index<0 || k_index>=int(one_ring.size()))
    {printf("Error in get_one_ring(%d,%d), size of one_ring=%d\n",k_polygon,k_edge,one_ring.size());exit(-1);}
  return one_ring[k_index];
}
int_vector Connectivity::get_one_ring(int k_vertex) const
{
  if(k_vertex<0 || k_vertex>=int(one_ring.size()))
    {printf("Error in get_one_ring(%d), size of one_ring=%d\n",k_vertex,one_ring.size());exit(-1);}
  return one_ring[k_vertex];
}

int Connectivity::get_polygon_access(int k_polygon) const
{
  if(k_polygon<0 || k_polygon>polygon_number())
    {printf("Error in Connectivity in get_polygon_access(%d), polygon_number()=%d\n",k_polygon,polygon_number());exit(-1);}
  return polygon_access[k_polygon];
}

int_vector Connectivity::get_polygon_ring(int k_polygon,int k_edge) const
{
  int k_index = get_index(k_polygon,k_edge);
  if(k_index<0 || k_index>=int(polygon_ring.size()))
    {printf("Error in get_polygon_ring(%d,%d), size of polygon_ring=%d\n",k_polygon,k_edge,polygon_ring.size());exit(-1);}
  return polygon_ring[k_index];
}
int_vector Connectivity::get_polygon_ring(int k_vertex) const
{
  if(k_vertex<0 || k_vertex>=int(polygon_ring.size()))
    {printf("Error in get_polygon_ring(%d), size of polygon_ring=%d\n",k_vertex,polygon_ring.size());exit(-1);}
  return polygon_ring[k_vertex];
}




int Connectivity::flip_polygon(int k_polygon)
{

  if(k_polygon<0 || k_polygon>=polygon_number())
    {printf("Error in flip_polygon(%d), N_polygon=%d\n",k_polygon,polygon_number());exit(-1);}
  
  int index=polygon_access[k_polygon];
  int size_current_polygon=size_polygon[k_polygon];
  const std::vector <int> current_polygon=get_index_of_polygon(k_polygon);
  if(int(current_polygon.size())!=size_current_polygon)
    {printf("Error in flip_polygon, size are not correct\n");exit(-1);}

  index_polygon[index]=current_polygon[0];
  int k_size_polygon=0;

  for(k_size_polygon=1;k_size_polygon<size_current_polygon;k_size_polygon++)
    index_polygon[index+k_size_polygon]=current_polygon[size_current_polygon-k_size_polygon];//in the other direction


  return 0;
}

int Connectivity::flip_polygon()
{
  int k=0;
  for(k=0;k<polygon_number();k++)
    flip_polygon(k);
  return 0;
}



int Connectivity::load_obj_file(const char* filename)
{std::string file=filename;return load_obj_file(file);}
int Connectivity::load_obj_file(const std::string& filename)
{
  destroy();

#define SIZE_BUFFER 2048
  char buffer[SIZE_BUFFER]={'\0'};
  char buffer2[SIZE_BUFFER]={'\0'};

  FILE *fid=NULL;
  fid=fopen(filename.data(),"r");
  if(fid==NULL)
    {printf("ERROR loading %s in load_obj_file in Connectivity\n",filename.data());exit(-1);}


  std::vector <int> poly;
  int cv=0;
  int one=-1;
  while(fscanf(fid,"%s",buffer)!=EOF)
    {
      switch(buffer[0])
	{
	case 'f': //polygon faces
	  poly.resize(0);
	  strcpy(buffer2,"\0");
	  fscanf(fid,"%s",buffer2);

	  if(strstr(buffer2,"//")!=NULL)
	    {
	      if(sscanf(buffer2,"%d//%*d",&cv)!=1)
		{printf("Error reading connectivity in %s\n",filename.data());exit(-1);}
	      poly.push_back(cv+one);
	      while(fscanf(fid,"%d//%*d",&cv)==1)
		poly.push_back(cv+one);
	      add_polygon(poly);
	    }
	  else if(sscanf(buffer2,"%d/%*d/%*d",&cv)==1)//warning %d/%d passes through it
	    {
	      poly.push_back(cv+one);
	      while(fscanf(fid,"%d/%*d/%*d",&cv)==1)
		poly.push_back(cv+one);
	      add_polygon(poly);
	    }
	  else if(sscanf(buffer2,"%d/%*d",&cv)==1)
	    {
	      poly.push_back(cv+one);
	      while(fscanf(fid,"%d/%*d",&cv)==1)
		poly.push_back(cv+one);
	      add_polygon(poly);
	    }
	  else if(sscanf(buffer2,"%d",&cv)==1)
	    {
	      poly.push_back(cv+one);
	      while(fscanf(fid,"%d",&cv)==1)
		poly.push_back(cv+one);
	      add_polygon(poly);
	    }
	  else //read the whole line
	    {
	      fgets(buffer,sizeof(buffer),fid);
	      //printf("Something wrong reading obj file [%s] at part:
	      //%s %s\n",filename.data(),buffer,buffer2);exit(-1);}
	    }
	  break;
	}
    }


  //build the index
  build_all_index();

  return 0;
}


int Connectivity::write_obj_type(const char* filename) const
{std::string file=filename;return write_obj_type(file);}
int Connectivity::write_obj_type(const std::string& filename) const
{
  FILE *fid=NULL;
  fid=fopen(filename.data(),"a");
  if(fid==NULL)
    {printf("ERROR loading %s in write_obj_type in Connectivity\n",filename.data());exit(-1);}
  
  int k_polygon=0,N_polygon=polygon_number();
  int current_size=0,k_index;
  int one=+1;
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    {
      current_size=polygon_size(k_polygon);

      //write the type of face
      fprintf(fid,"f ");

      //read the polygon
      for(k_index=0;k_index<current_size;k_index++)
	fprintf(fid,"%d ",get_index(k_polygon,k_index)+one);

      fprintf(fid,"\n");
    }

  if(fclose(fid)!=0)
    {printf("ERROR closing %s in write_obj_type in Connectivity\n",filename.data());exit(-1);}


  return 0;
}

int Connectivity::triangulate()
{
  if(triangulated_index.size()%3!=0)
    {printf("Error in triangulate() in Connectivity, size of triangulated_index is not multiple of 3 (%d)\n",triangulated_index.size());exit(-1);}

  int k_vertex=0,N_vertex=triangulated_index.size()/3;
  index_polygon.resize(0);
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    add_polygon(triangulated_index[3*k_vertex+0],triangulated_index[3*k_vertex+1],triangulated_index[3*k_vertex+2]);
  return build_all_index();
}






int Connectivity::load_g_file(const char* filename)
{return load_g_file(std::string(filename));}
int Connectivity::load_g_file(const std::string& filename)
{
    destroy();

  #define SIZE_BUFFER 2048
  char buffer[SIZE_BUFFER]={'\0'};

  FILE *fid=NULL;
  fid=fopen(filename.data(),"r");
  if(fid==NULL)
    {printf("ERROR loading %s in load_g_file in Connectivity\n",filename.data());exit(-1);}


  //pass comments
  fscanf(fid,"%s",buffer);
  while(buffer[0]=='#')//comments
    {
      fgets(buffer,sizeof(buffer),fid);
      fscanf(fid,"%s",buffer);
    }

  //read Number of polygon
  int N_vertex=0,N_polygon=0;
  N_vertex=atoi(buffer);//N_vertex
  fscanf(fid,"%d",&N_polygon);//N_polygon
  fgets(buffer,sizeof(buffer),fid);//read line

  //read vertices
  int k_vertex=0;
  for(k_vertex=0;k_vertex<N_vertex;k_vertex++)
    {
      fgets(buffer,sizeof(buffer),fid);//read line
      if(buffer[0]=='#')//comments does not count in vertices number
	k_vertex--;
    }

  //read polygons

  int size_poly=0;
  int k_polygon=0;
  int k_current_polygon=0;
  int temp=0;
  std::vector <int> v_poly;
  for(k_polygon=0;k_polygon<N_polygon;k_polygon++)
    {
       if(fscanf(fid,"%d",&size_poly)!=1)
	{
	  fscanf(fid,"%s",buffer);
	  if(buffer[0]!='#')
	    {printf("ERROR loading polygons in load_g_file %s in Connectivity\n",filename.data());exit(-1);}
	  else//read line
	    {
	      fgets(buffer,sizeof(buffer),fid);
	      k_polygon--;
	    }
	}
      else
	{
	  v_poly.resize(0);
	  for(k_current_polygon=0;k_current_polygon<size_poly;k_current_polygon++)
	    {
	      fscanf(fid,"%d",&temp);
	      v_poly.push_back(temp);
	    }
	  add_polygon(v_poly);
	  //read rest of the line
	  fgets(buffer,sizeof(buffer),fid);
	}
    }

  if(fclose(fid)!=0)
    {printf("ERROR closing %s in load_g_file in Connectivity\n",filename.data());exit(-1);}

  //build the index
  build_all_index();

  return 0;


}


