//** Modified-Light version of V_4D for TP**// 


#include <V_4D.h>

V_4D::V_4D(){for(int k_dim=0;k_dim<4;k_dim++)V[k_dim]=0.0;}
V_4D::V_4D(const V_4D &v){for(int k_dim=0;k_dim<4;k_dim++)V[k_dim]=v[k_dim];}
V_4D::V_4D(double x,double y,double z,double w){V[0]=x;V[1]=y;V[2]=z;V[3]=w;}
V_4D::~V_4D(){}

double& V_4D::operator[](int index)
{
  if(index<0 || index>3)
    {printf("Error in V_4D[%d], index must be between [0,3]\n",index);exit(-1);}
  return V[index];
}
double V_4D::operator[](int index) const
{
  if(index<0 || index>3)
    {printf("Error in V_4D[%d], index must be between [0,3]\n",index);exit(-1);}
  return V[index];
}
double& V_4D::operator()(int index)
{
  if(index<0 || index>3)
    {printf("Error in V_4D(%d), index must be between [0,3]\n",index);exit(-1);}
  return V[index];
}
double V_4D::operator()(int index) const
{
  if(index<0 || index>3)
    {printf("Error in V_4D(%d), index must be between [0,3]\n",index);exit(-1);}
  return V[index];
}



int V_4D::equal(const V_4D& v,double epsilon) const
{
  if(pow((V[0]-v[0])*(V[0]-v[0])+(V[1]-v[1])*(V[1]-v[1])+(V[2]-v[2])*(V[2]-v[2])+(V[3]-v[3])*(V[3]-v[3]),0.5)<=epsilon)
    return 1;
  else
    return 0;
}
int V_4D::equal(const V_4D& v) const
{
  double epsilon=0.00001;
  return equal(v,epsilon);
}

V_4D& V_4D::operator=(const V_4D& _v)
{
  for(int k_dim=0;k_dim<4;k_dim++)
    V[k_dim] = _v.V[k_dim];
  return *this;
}

V_4D& V_4D::operator+=(const V_4D& v)
{
  for(int k_dim=0;k_dim<4;k_dim++)
    V[k_dim]+=v[k_dim];
  return *this;
}
  
V_4D& V_4D::operator/=(const double& alpha)
{
  for(int k_dim=0;k_dim<4;k_dim++)
    V[k_dim]/=alpha;
  return *this;
}


V_4D operator+(const V_4D& v1,const V_4D& v2)
{
  V_4D Z;
  for(int k_dim=0;k_dim<4;k_dim++)
    Z[k_dim]=v1[k_dim]+v2[k_dim];
  return Z;
}
V_4D operator-(const V_4D& v1,const V_4D& v2)
{
  V_4D Z;
  for(int k_dim=0;k_dim<4;k_dim++)
    Z[k_dim]=v1[k_dim]-v2[k_dim];
  return Z;
}
V_4D operator*(const V_4D& v1,const double& alpha)
{
  V_4D Z;
  for(int k_dim=0;k_dim<4;k_dim++)
    Z[k_dim]=alpha*v1[k_dim];
  return Z;
}
V_4D operator*(const double& alpha,const V_4D& v1)
{
  V_4D Z;
  for(int k_dim=0;k_dim<4;k_dim++)
    Z[k_dim]=alpha*v1[k_dim];
  return Z;
}
V_4D operator/(const V_4D& v1,const double& alpha)
{
  V_4D Z;
  for(int k_dim=0;k_dim<4;k_dim++)
    Z[k_dim]=v1[k_dim]/alpha;
  return Z;
}



bool operator==(const V_4D& v1,const V_4D& v2)
{return (v1.equal(v2)==1);}
bool operator!=(const V_4D& v1,const V_4D& v2)
{return (v1.equal(v2)==0);}

std::ostream& operator << (std::ostream& flux, V_4D _v)
{
  flux<<"("<<_v[0]<<","<<_v[1]<<","<<_v[2]<<","<<_v[3]<<")";
  return flux;
}

int V_4D::set_zero(){V[0]=0;V[1]=0;V[2]=0;V[3]=0; return 0;}
int V_4D::set_3d_zero(){V[0]=0;V[1]=0;V[2]=0;V[3]=1;return 0;}
