
//** TP VM - RICM3 **/


//libs
#include <stdlib.h>
#include <iostream>
#include <cmath>
#include <string>
#include <sstream>
#include <errno.h>
#include <string>
#include <stdio.h>
#include <vector>
using namespace std;


//openGL
#include <GL/glut.h>

//class
#include <Mesh_object.h>
#include <Mesh_object_loader.h>
#include <Texture.h>
#include <OpenGL_drawer.h>
#include <Navigator_tool.h>


#define PI 3.14159



//************************************//
/** GLOBAL FUNCTION **/
//************************************//



//glut functions
static void display_callback();
static void reshape_callback (int width, int height);
static void keyboard_callback(unsigned char key,int x,int y);
static void idle_callback(int value);
static void motion_callback (int x, int y);
static void mouse_callback (int button, int action, int x, int y);


//main functions
int setup_opengl(int argc,char **argv);
int draw_scene(); //draw the meshes
int draw_skybox(); //draw the skybox
int load_model(); //load the meshes
int load_skybox(); //load the skybox

// some helpers
double get_rand();//just a rand
double get_centered_rand();//just a centered rand




//************************************//
//GLOBAL VARIABLES
//************************************//

// helper class to draw mesh
OpenGL_drawer openGL_drawer;
// helper class to store information about camera, windows, ...
Navigator_tool navigator;

// a temp mesh
Mesh_object mesh;


// the frame counter
int counter=0;
// the number of polygon drawn on screen
int N_polygon=0;

// Number of meshes (can change this value)
int N_mesh = 40;
// Number of subdivision level (see you data/)
int N_subdiv = 8;

// stored vectors
std::vector <Mesh_object> v_mesh; // store the multires-meshes
std::vector <V_3D> v_position; // position of each mesh
std::vector <V_3D> v_speed; // speed of each mesh
std::vector <V_3D> v_acceleration; //acceleration of each mesh
std::vector <int> v_subdivision_number; // subdivision level of each mesh

// z distance for which we change of subdivision level
std::vector <double> threshold_distance; 


// the skybox
Mesh_object skybox_left,skybox_right,skybox_forward,skybox_backward,skybox_down,skybox_up;
double Lx=115*2.0,Ly=6.0*2; //size of skybox

// counter for walking
int walk_count=0;

// should we draw the grid (keyboard: D)
int is_draw_grid=0;
// should we draw the normal per vertex/polygon  (keyboard: N)
int is_normal_polygon=0;




int main(int argc, char *argv[])
{
  cout<<"** TP VM RICM-3 **\n"<<endl;
  printf("run %s with %d parameters ... \n\n",argv[0],argc-1);


  //start OpenGL
  //*************//
  navigator.set_screen_size(800,800);
  setup_opengl(argc,argv);



  //load  models
  //*************//
  load_model();
  load_skybox();



  // init some vectors
  for(int k=0;k<N_mesh;k++)
    {
      v_subdivision_number.push_back(0);
      v_position.push_back(V_3D(10*get_centered_rand(),0.0,10*get_centered_rand()));
      v_speed.push_back(V_3D(get_centered_rand(),0.0,get_centered_rand()));
    }
  v_acceleration.resize(N_mesh);


  // the threshold limit for the subdivision level could be changed by hand
  threshold_distance.resize(N_subdiv-1);
  for(int k_subdiv=0;k_subdiv<N_subdiv-1;k_subdiv++)
    threshold_distance[k_subdiv]=k_subdiv*2;



  //run
  printf("**RUN**\n\n");
  printf("Navigation with keys:\n\n");
  printf("==========================\n");
  printf("Space Displacement\n");
  printf("  z :Forward  \n");
  printf("  s :Backward \n\n");
  printf("and mouse with drag & drop\n\n");
  printf("W: wireframe mode\n");
  printf("F: filled mode\n");
  printf("N: Normal type per vertex/polygon\n");
  printf("D: Grid on/off\n");
  printf("==========================\n");
  glutMainLoop ();


  return 0;
}



//load the meshes
int load_model()
{

  // set the name of your model here (will fill with NAME_[i], where i varies with N_subdiv)
  std::string filename("data/meshes/dragon_"); //for dragon
  //std::string filename("data/meshes/dino_"); //for dino


  std::string extension;
  std::string full_name;
  std::ostringstream string_number("");

  // load each subdivision
  for(int k_mesh=0;k_mesh<N_subdiv;k_mesh++)
    {
      // reconstruct the name
      string_number.seekp(0);
      string_number<<k_mesh;
      extension=string_number.str();
      extension += ".off";
      full_name = filename+extension;

      // load it
      mesh.load_off_file( full_name );

      // scale/rotate ... the loaded mesh
      // ********************************* //
      mesh.scale(11.0);
      mesh.rotate(V_3D(0,1,0),PI);
      //mesh.scale(0.01); //for dino
      //mesh.rotate(V_3D(0,0,1),-PI/2.0); //for dino
      //mesh.rotate(V_3D(1,0,0),-PI/2.0); //for dino
      mesh.build_normal_per_vertex();

      // set the color of the meshes here
      mesh.fill_color(0.1,0.4,0.2,1.0);

      // make sure the mesh tuch the floor
      mesh.translate(-1.0*mesh.get_min_point());

      // set its name
      mesh.set_name(full_name);

      // add it into the storing vector
      v_mesh.push_back(mesh);

      // store a drawable-ready version in the GPU
      openGL_drawer.add_draw_list(mesh);
    }



  return 0;
}









//Draw the scene
int draw_scene()
{

  glEnable(GL_LIGHTING);

  //current position in space
  V_3D position = -1*navigator.get_current_position();

  // position of the camera
  Matrix camera;
  // get the screen camera orientation
  float float_camera[16];
  glGetFloatv(GL_MODELVIEW_MATRIX,float_camera); //get openGL matrix

  //openGL store its matrix in a transposed mode
  for(int k1=0;k1<4;k1++)
    for(int k2=0;k2<4;k2++)
      camera(k1,k2)=float_camera[k1+4*k2];
  
  //the transformation matrix for the meshes
  Matrix M;

  // the current level of subdivision
  int current_subdivision=1;


  /// frequency of the polygon counter
  int Polygon_counter_frequency=200;
  /// update the polygon counter
  if(counter%Polygon_counter_frequency==0)
    N_polygon=0;


  //******************//
  //drawing LOOP
  //******************//
  glEnable(GL_LIGHTING);
  for(int k_mesh=0;k_mesh<N_mesh;k_mesh++)
    {
      //set the transformation matrix for the current mesh
      M.build_rotation_matrix_to_vector(V_3D(1,0,0),v_speed[k_mesh]);
      M.set_translation(v_position[k_mesh]);
      M=M.transpose();

      // now rotate the matrices for each mesh
      glPushMatrix();
      {
	glMultMatrixd(M.get_pointer()); //rotation
	
	//draw the subdivision current_subdivision stored in a draw_list
	openGL_drawer.draw_in_list(current_subdivision);
	
	//if we want the grid
	if(is_draw_grid==1)
	  {
	    glLineWidth(1.0); glColor4d(0.1,0.1,0.8,1.0);
	    openGL_drawer.draw_mesh_grid(v_mesh[current_subdivision]);
	  }
	
	// update the polygon counter in some case
	if(counter%Polygon_counter_frequency==0)
	  N_polygon += v_mesh[current_subdivision].polygon_number();
      }
      glPopMatrix();
    }


  draw_skybox();


  return 0;
}

int draw_skybox()
{
  //just quads, so no light
  glDisable(GL_LIGHTING);
  glPushMatrix();
  {
    //follow the camera
    glTranslated(-1.0*navigator.get_tr_x(),-1.0*navigator.get_tr_y(),-1.0*navigator.get_tr_z());

    //draw the quads (the ground is special)
    openGL_drawer.draw_mesh_slow(skybox_forward);
    openGL_drawer.draw_mesh_slow(skybox_left);
    openGL_drawer.draw_mesh_slow(skybox_backward);
    openGL_drawer.draw_mesh_slow(skybox_right);
    openGL_drawer.draw_mesh_slow(skybox_up);
  }
  glPopMatrix();

  //now, draw the ground (we draw it many times, but it does not move with the camera)
  for(int kx=-2;kx<=2;kx++)
    for(int ky=-2;ky<=2;ky++)
      {
	glPushMatrix();
	glTranslated(kx*Lx,0,ky*Lx);
	openGL_drawer.draw_mesh_slow(skybox_down);
	glPopMatrix();
      }
  glEnable(GL_LIGHTING);

  return 0;
}





int load_skybox()
{
  // load the different skybox
  char quad_filename[500]="data/meshes/textured_quad.obj";

  Mesh_object_loader helper;
  skybox_forward.load_obj_file(quad_filename);
  helper.load_obj_texture(quad_filename,&skybox_forward);
  skybox_forward.fill_color(0.8,0.8,0.8,1.0);
  skybox_forward.set_texture_number(openGL_drawer.load_texture("data/textures/XN.ppm"));

  skybox_left.load_obj_file(quad_filename);
  skybox_left.rotate_centered(V_3D(0,1,0),PI);
  skybox_left.rotate(V_3D(0,1,0),90.0*PI/180.0);
  helper.load_obj_texture(quad_filename,&skybox_left);
  skybox_left.fill_color(0.8,0.8,0.8,1.0);
  skybox_left.set_texture_number(openGL_drawer.load_texture("data/textures/ZP.ppm"));

  skybox_right.load_obj_file(quad_filename);
  skybox_right.rotate(V_3D(0,1,0),90.0*PI/180.0);
  skybox_right.translate(V_3D(1,0,0));  
  helper.load_obj_texture(quad_filename,&skybox_right);
  skybox_right.fill_color(0.8,0.8,0.8,1.0);
  skybox_right.set_texture_number(openGL_drawer.load_texture("data/textures/ZN.ppm"));

  skybox_backward.load_obj_file(quad_filename);
  helper.load_obj_texture(quad_filename,&skybox_backward);
  skybox_backward.translate(V_3D(0,0,-1));  
  skybox_backward.fill_color(0.8,0.8,0.8,1.0);
  skybox_backward.set_texture_number(openGL_drawer.load_texture("data/textures/XP.ppm"));

  skybox_down.load_obj_file(quad_filename);
  helper.load_obj_texture(quad_filename,&skybox_down);
  skybox_down.rotate(V_3D(1,0,0),PI/2.0);  
  skybox_down.fill_color(0.8,0.8,0.8,1.0);
  skybox_down.set_texture_number(openGL_drawer.load_texture("data/textures/YN.ppm"));
  
  skybox_up.load_obj_file(quad_filename);
  helper.load_obj_texture(quad_filename,&skybox_up);
  skybox_up.rotate(V_3D(1,0,0),PI/2.0);  
  skybox_up.fill_color(0.8,0.8,0.8,1.0);
  skybox_up.set_texture_number(openGL_drawer.load_texture("data/textures/YP.ppm"));
  

  // scale and translate it correctly

  double ty=0.01;
  skybox_forward.translate(V_3D(-0.5,-ty,+0.5));
  skybox_backward.translate(V_3D(-0.5,-ty,+0.5));
  skybox_right.translate(V_3D(-0.5,-ty,+0.5));
  skybox_left.translate(V_3D(-0.5,-ty,+0.5));



  skybox_forward.scale(Lx,Ly,Lx);
  skybox_backward.scale(Lx,Ly,Lx);
  skybox_right.scale(Lx,Ly,Lx);
  skybox_left.scale(Lx,Ly,Lx);

  skybox_down.translate(-0.5,0,-0.5);  
  skybox_down.scale(Lx,1.0,Lx);
  skybox_down.translate(V_3D(0.0,-ty*Ly,0.0));

  skybox_up.translate(-0.5,0,-0.5);  
  skybox_up.scale(Lx,1.0,Lx);
  skybox_up.translate(V_3D(0.0,(1.0-ty)*Ly,0.0));
  
  return 0;
}







//*******************//
//GLUT STUFF
//*******************//








//glut functions
int setup_opengl(int argc,char **argv)
{
  openGL_drawer.init_opengl(argc,argv,navigator);

  //callback
  glutDisplayFunc (display_callback);
  glutReshapeFunc (reshape_callback);
  glutKeyboardFunc(keyboard_callback);
  glutTimerFunc(33,idle_callback,1);
  glutMotionFunc (motion_callback);
  glutMouseFunc (mouse_callback);

  return 0;
}



//display
static void display_callback()
{
  //croping
  double Z_MIN = 0.05;
  double Z_MAX = 520.0;


  //vieport (all screen)
  glViewport (0, 0, navigator.get_X_screen(),navigator.get_Y_screen());
  glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  //Projection
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity ();
  gluPerspective (45, navigator.get_screen_ratio(), Z_MIN,Z_MAX);
  glTranslated(0,-1.0,0);

  //Modelview
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // transformation of the camera
  glRotated(navigator.get_theta_2(),0,1,0);
  glTranslatef(GLfloat(navigator.get_tr_x()),GLfloat(navigator.get_tr_y()),GLfloat(navigator.get_tr_z())); //translation



  //Draw the scene
  draw_scene();

    
  //update fps
  navigator.set_current_time(glutGet(GLUT_ELAPSED_TIME));
  if(navigator.get_delta_time()>1000)
    navigator.refresh_fps();

  //draw fps
  glColor4d(0.0,0.0,0.0,1.0);
  openGL_drawer.draw_fps(navigator);

  //write on the screen the number of polygons drown
  glDisable(GL_DEPTH_TEST);
  glPushMatrix();
  glLoadIdentity();
  char strpoly[50]="\0";
  glRasterPos3f(-0.52f,+1.45f,-1.3f);
  sprintf ( strpoly, "%d Polygons", N_polygon );
  for (unsigned int i=0;i<strlen(strpoly);i++) 
    glutBitmapCharacter(GLUT_BITMAP_TIMES_ROMAN_24,*(strpoly+i));
  glPopMatrix();
  glEnable(GL_DEPTH_TEST);



  //Error code in OpenGL
  GLenum errCode;
  const GLubyte *errString;
  if((errCode=glGetError()) != GL_NO_ERROR)
    {
      errString = gluErrorString(errCode);
      printf("ERROR OPENGL : %s\n",errString);
      exit(-2);
    }

  //redraw
  glFlush ();
  glutSwapBuffers ();

  // update some counters
  counter++;
  navigator.increase_frame();

}


//reshape the screen
static void reshape_callback (int width, int height)
{
  navigator.set_screen_size(width,height);
}


//keyboard
static void keyboard_callback(unsigned char key,int x,int y)
{

  x=0;y=0;
  double d_L=0.001*(1+100*abs(navigator.get_dist_obs()));

  #define GLUT_SPACE 32

  V_3D p(navigator.get_tr_x(),navigator.get_tr_y(),navigator.get_tr_z());
  switch(key)
    {
    case 'Q'://quit
      printf("\n[EXIT OK]\n\n");
      exit(0);
      break;

      //Displacement
    case 'z': //forward
      navigator.go_forward(d_L);
      navigator.go_up(0.01*cos(2*PI*double(walk_count)/40.0)+0.0025*cos(6.5*PI*double(walk_count)/70.0));
      walk_count++;
      break;
    case 's': //backward
      navigator.go_backward(d_L);
      navigator.go_up(0.01*cos(2*PI*double(walk_count)/40.0)+0.0025*cos(6.5*PI*double(walk_count)/70.0));
      walk_count++;
      break;


    case 'W'://wireframe
      glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
      break;
    case 'F'://Filled
      glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);
      break;
    case 'D'://draw_grid
      is_draw_grid==0?is_draw_grid=1:is_draw_grid=0;
    break;
    case 'N'://normal_mode
      {
	is_normal_polygon==0?is_normal_polygon=1:is_normal_polygon=0;
	if(is_normal_polygon==1)
	  {
	    for(int k_mesh=0;k_mesh<N_subdiv;k_mesh++)
	      {
		v_mesh[k_mesh].build_normal_per_polygon();
		openGL_drawer.update_draw_list(v_mesh[k_mesh]);
	      }
	  }
	else
	  {
	    for(int k_mesh=0;k_mesh<N_subdiv;k_mesh++)
	      {
		v_mesh[k_mesh].build_normal_per_vertex();
		openGL_drawer.update_draw_list(v_mesh[k_mesh]);
	      }
	  }

      }
      break;
    }


  glutPostRedisplay ();
}


static void idle_callback(int value)
{
  value=1;//useless

  // time integration step
  double d_t=0.01;
  // the radius of the group
  double L_max=25;

  // some temp position
  V_3D v0,v1;
  // the current position of the camera
  V_3D current_position=V_3D(-navigator.get_tr_x(),0.0,-navigator.get_tr_z());




  //** LOOP OTHER Every mesh to update there speed, ... **//

  for(int k_mesh=0;k_mesh<N_mesh;k_mesh++)
    {
      // integration with damping
      v_speed[k_mesh] = v_speed[k_mesh] + d_t*v_acceleration[k_mesh];
      v_position[k_mesh] = v_position[k_mesh] + d_t*v_speed[k_mesh]*(1-0.1);
      v_acceleration[k_mesh] = v_acceleration[k_mesh]*0.9; //(to avoir fast direction change)
    }
  for(int k_mesh=0;k_mesh<N_mesh;k_mesh++)
    {
      // sometimes, can change of direction
      if(counter%25==0 && get_rand()>0.7)
	v_acceleration[k_mesh] = v_acceleration[k_mesh]/2.0+ V_3D(get_centered_rand(),0.0,get_centered_rand());

      // Try to avoid each other (do not prevent collision at all!)
      for(int k_mesh2=0;k_mesh2<N_mesh;k_mesh2++)
	if(k_mesh2!=k_mesh)
	  {
	    v0=v_position[k_mesh];
	    v1=v_position[k_mesh2];
	    if( (v0-v1).norm()<6 )
	      v_acceleration[k_mesh] =  (v0-v1).normalized();
	  }

      // follow the camera in a certain radius      
      if((v_position[k_mesh]-current_position).norm()>L_max)
	v_acceleration[k_mesh] = -0.5*(v_position[k_mesh]-current_position).normalized()*2*get_rand();
    }

  // update
  navigator.increase_time();
  glutPostRedisplay ();
  glutTimerFunc( 10, idle_callback, 1);

}



// when we move the mouse
static void motion_callback (int x, int y)
{
  /* left button */
  if (navigator.get_left_b()) 
    {
      navigator.set_theta_2(navigator.get_theta_2()+0.25*double(x-navigator.get_x_prec()));
      navigator.set_x_prec(x);
      navigator.set_y_prec(y);
    }

  glutPostRedisplay ();
}
// when we click on the mouse
static void mouse_callback (int button, int action, int x, int y)
{

  if (button == GLUT_LEFT_BUTTON && action == GLUT_DOWN) {
    navigator.set_left_b(1);
    navigator.set_x_prec(x); 
    navigator.set_y_prec(y);
  }
  if (button == GLUT_LEFT_BUTTON && action == GLUT_UP) 
    navigator.set_left_b(0); 

  if (button == GLUT_RIGHT_BUTTON && action == GLUT_DOWN) {
    navigator.set_right_b(1);
    navigator.set_y_prec(y);
  }
  if (button == GLUT_RIGHT_BUTTON && action == GLUT_UP) 
    navigator.set_right_b(0); 

  glutPostRedisplay ();

}


//just a rand number \in [0,1]
double get_rand()
{
  int precision=50000;
  return double(rand()%precision)/double(precision);
}

//just a rand number \in [-1,1]
double get_centered_rand()
{return 2*get_rand()-1;}


